/*
 * Copyright (C) 2012 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.sdkuilib.internal.repository.ui;


/**
 * Icons used by {@link PackagesPage}.
 */
public class PackagesPageIcons {

    public static final String ICON_CAT_OTHER      = "pkgcat_other_16.png";    //$NON-NLS-1$
    public static final String ICON_CAT_PLATFORM   = "pkgcat_16.png";          //$NON-NLS-1$
    public static final String ICON_SORT_BY_SOURCE = "source_icon16.png";      //$NON-NLS-1$
    public static final String ICON_SORT_BY_API    = "platform_pkg_16.png";    //$NON-NLS-1$
    public static final String ICON_PKG_NEW        = "pkg_new_16.png";         //$NON-NLS-1$
    public static final String ICON_PKG_INCOMPAT   = "pkg_incompat_16.png";    //$NON-NLS-1$
    public static final String ICON_PKG_UPDATE     = "pkg_update_16.png";      //$NON-NLS-1$
    public static final String ICON_PKG_INSTALLED  = "pkg_installed_16.png";   //$NON-NLS-1$
}
