/*
 * Copyright (C) 2010 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.ddmlib;


/**
 * Exception thrown when a shell command executed on a device takes too long to send its output.
 * <p/>The command may not actually be unresponsive, it just has spent too much time not outputting
 * any thing to the console.
 */
public class ShellCommandUnresponsiveException extends Exception {
    private static final long serialVersionUID = 1L;
}
